﻿using Upreader.Common;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using System.IO;

namespace Upreader.Application.Tests
{
    class ChunkedMemoryStreamStub : ChunkedMemoryStream
    {
        public ChunkedMemoryStreamStub(int capacity)
            : base(capacity)
        {

        }

        public void WriteStringLine(string input)
        {
            MemoryStream bytesStream = new MemoryStream();
            StreamWriter writer = new StreamWriter(bytesStream);
            writer.WriteLine(input);
            writer.Flush();

            WriteMergedBytes(bytesStream.ToArray());
        }

        public string ReadStringLine()
        {
            MemoryStream bytesStream = new MemoryStream(ReadMergedBytes());
            StreamReader reader = new StreamReader(bytesStream);
            return reader.ReadLine();
        }

        public void WriteMergedBytes(byte[] bytes)
        {
            PrepareCapacity(bytes.Length);

            int totalChunks = Chunks.Count();
            for (int index = 0; index < totalChunks; index++)
            {
                var currentSeries = bytes.Skip(index * ChunkSize).Take(ChunkSize);
                currentSeries.ToArray().CopyTo(Chunks[index], 0);
            }
        }

        public byte[] ReadMergedBytes()
        {
            return Chunks.SelectMany(x => x).ToArray();
        }
    }

    [TestClass]
    public class ChunkedMemoryStreamTest
    {
        [TestMethod]
        public void Read_MultipleChunks_Succeeds()
        {
            // arrange
            string inputLine = "Hello world spanning multiple chunks";
            ChunkedMemoryStreamStub target = new ChunkedMemoryStreamStub(10);
            target.WriteStringLine(inputLine);

            // act
            StreamReader reader = new StreamReader(target);
            string line = reader.ReadLine();

            // assert
            Assert.AreEqual(line, inputLine);
        }

        [TestMethod]
        public void Write_MultipleChunks_Succeeds()
        {
            // arrange
            string inputLine = "Hello world spanning multiple chunks";
            ChunkedMemoryStreamStub target = new ChunkedMemoryStreamStub(10);
            StreamWriter writer = new StreamWriter(target);

            // act
            writer.WriteLine(inputLine);
            writer.Flush();

            // assert
            string line = target.ReadStringLine();
            Assert.AreEqual(line, inputLine);
        }

        [TestMethod]
        public void Length_AfterWrite_Succeeds()
        {
            // arrange
            string inputLine = "Hello world spanning multiple chunks";
            ChunkedMemoryStreamStub target = new ChunkedMemoryStreamStub(10);
            StreamWriter writer = new StreamWriter(target);

            // act
            writer.Write(inputLine);
            writer.Flush();
            int total = (int)target.Length;

            // assert
            Assert.AreEqual(total, inputLine.Length);
        }
    }
}
